# -*- coding: utf-8 -*-

import os
import sys

import typer
from typing_extensions import Annotated
from . import download as dl
from . import get_seq as get
from . import infer_allele
from . import __version__
from . import hladlfunctions as fxn

# Ensure correct importlib-resources function imported
if sys.version_info < (3, 9):
    import importlib_resources                              # PyPI
else:
    import importlib.resources as importlib_resources       # importlib.resources


app = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]})


data_files = importlib_resources.files("hladldata")
data_dir = os.path.dirname(str(data_files / '__init__.py'))


@app.callback()
def callback():
    """
    hladl: a command line tool for getting HLA sequences
    """


@app.command()
def init(seqtype: Annotated[str, typer.Option("--seqtype", "-s",
            help="Type of sequence to download: gen, nuc, prot, or all.")] = 'all',
         digits: Annotated[int, typer.Option("--digits", "-d",
            help="Digit length to download at: 2, 4, 6.")] = 4,
         ):
    """
    Initialise the tool by automatically downloading the HLA data from https://github.com/ANHIG/IMGTHLA
    """

    dl.get_data(seqtype, digits, data_dir)


@app.command()
def seq(allele: Annotated[str, typer.Option("--allele", "-a",
                  help="Allele to download, in the format [ABC]*XX:XX, with the appropriate number of digits.")],
        digits: Annotated[int, typer.Option("--digits", "-d",
                  help="Digit length to download at: 2, 4, 6.")] = 4,
        seqtype: Annotated[str, typer.Option("--seqtype", "-s",
                  help="Type of sequence to download: nuc or prot")] = 'prot',  # TODO add gen?
        mode: Annotated[str, typer.Option('-m', '--mode',
                  help="hladl mode: full (regular), ecd (extracellular domain")] = 'full',
        output_mode: Annotated[str, typer.Option('-om', '--output_mode',
                  help="Output mode: stdout (terminal), fa (FASTA)")] = 'stdout',
        ):
    """
    Output the sequence of a specified HLA allele
    """

    hla_seq = get.seq(allele, digits, seqtype, mode, output_mode, data_dir)
    if output_mode == 'stdout':
        if hla_seq:
            print(hla_seq)
        else:
            print("Failed to grab the", seqtype, "sequence for", allele, "at", str(digits), "resolution. ")



@app.command()
def infer(seq: str,
          digits: Annotated[int, typer.Option("--digits", "-d",
                  help="Digit length to download at: 2, 4, 6.")] = 4,
          seqtype: Annotated[str, typer.Option("--seqtype", "-s",
                  help="Type of sequence to download: nuc or prot")] = 'prot',
          # TODO add gen?
          hla_loci: Annotated[str, typer.Option("--hla_loci", "-L",
                help='Comma-delimited list of IMGTHLA-formatted gene prefixes'
                     ' (before the *)')] = ','.join(fxn.featured_mhci),
          output_mode: Annotated[str, typer.Option("--output_mode", "-om",
                 help="Output mode: stdout (terminal), fa (FASTA)")] = 'stdout',
          lower_len: Annotated[int, typer.Option("--lower_len", "-l",
                 help="Lower length limit of allele product (amino acids) to filter. 0 = no filter.")] = 350,
          upper_len: Annotated[int, typer.Option("--upper_len", "-u",
                 help="Upper length limit of allele product (amino acids) to filter. 0 = no filter.")] = 380,
          window_len: Annotated[int, typer.Option("--window_len", "-wl",
                 help="Sliding window length .")] = 20):
    """
    Given an HLA sequence, infer what allele(s) it could be derived from
    """

    inferred_alleles, tag_count = infer_allele.infer(seq, digits, seqtype, hla_loci, data_dir, lower_len, upper_len, window_len)

    if output_mode.lower() == 'stdout':
        print(f"Detected top-matching alleles: {inferred_alleles}")
        print(f"Number of tags per hit: {tag_count}")



@app.command()
def dd():
    """
    Print the location of the hladl data directory
    """
    print(data_dir)
    return data_dir


@app.command()
def version():
    """
    Print the hladl package version
    """
    print(__version__)



# -*- coding: utf-8 -*-

from acora import AcoraBuilder
import collections as coll
import os
from . import hladlfunctions as fxn


def build_trie(sequence_list):
    """
    :param sequence_list: A list of DNA sequences (tags) to search for
    :return: An Aho-Corasick trie, which can be used to search sequences downstream with all of the listed tags
    """

    trie_builder = AcoraBuilder()
    for sequence in sequence_list:
        trie_builder.add(sequence)

    return trie_builder.build()


def get_trie_data(num_digits, seq_type, hla_dir, gene_prefixes, lower, higher, slide_len):
    """
    Generate the necessary tag files for easy loading for AC trie generation,
        saving to or loading from hladl data dir as needed
    :param num_digits: int, resolution of HLA allele number (2/4/6/8)
    :param seq_type: str, sequence type as per IMGTHLA (gen, nuc, prot)
    :param hla_dir: str, path to hladl data directory
    :param gene_prefixes: list of str, detailing IMGTLA pre-asterisk gene identifiers
    :param lower: int, lower length limit of sequence allowed (in amino acids)
    :param higher: int, upper length limit of sequence allowed (in amino acids)
    :param slide_len: int, length of sliding window to use for overlapping tags
    :return: AC trie generated from tags built from the appropriate data set
    """
    hla, data_file = fxn.get_data(num_digits, seq_type, hla_dir)

    # If the tags exist, read them in...
    tag_file = os.path.join(hla_dir, data_file[:data_file.index('.json')] + '_tags.json.gz')

    if os.path.isfile(tag_file):
        tags2hla = fxn.read_json(tag_file)

    # ... Otherwise, let's make a fresh set
    else:
        kept_alleles = coll.defaultdict(str)

        if lower == 0 and higher == 0:
            len_check = False
        else:
            len_check = True
            if seq_type == 'nuc':
                lower *= 3
                higher *= 3

        # Filter down to just the wanted MHC genes / lengths
        for allele in hla:

            if allele.split('*')[0] not in gene_prefixes:
                continue

            if len_check:
                len_seq = len(hla[allele])
                if lower:
                    if len_seq < lower:
                        continue
                if higher:
                    if len_seq > higher:
                        continue

            # Otherwise keep and use
            kept_alleles[allele] = hla[allele]

        # Then make sliding windows
        tags2hla = coll.defaultdict(list)
        for allele in kept_alleles:
            hla_seq = kept_alleles[allele]

            # Take sliding windows across the full length genes, overlapping by half their length
            for i in range(0, len(hla_seq) - slide_len + 1, int(slide_len / 2)):
                slide = hla_seq[i:i + slide_len]
                if len(slide) == slide_len:
                    tags2hla[slide].append(allele)

        # Then save that (to speed up repeat inferences with the same parameters
        fxn.save_json(tag_file, tags2hla)

    # Return a trie of that, and the
    return build_trie(list(tags2hla.keys())), tags2hla


def infer_hla_allele(pot_hla_seq, search_trie, tags2hla):
    """
    :param pot_hla_seq: str, nucleotide or amino acid sequence that (maybe) contains an HLA seq
    :param search_trie: Aho-Corasick object, detailing the appropriate trie to search HLA strings
    :param tags2hla: dict describing which HLA genes are covered by a specific tag sequence
    :return: 2 items: list of str detailing top scoring alleles (sorted alphabetically), & int of their  tag count
    """

    pot_hla_seq = pot_hla_seq.upper()
    matches = search_trie.findall(pot_hla_seq)
    if matches:
        all_matching_alleles = coll.Counter(flatten([tags2hla[x[0]] for x in matches])).most_common()
        highest_score = all_matching_alleles[0][1]
        most_common = [x for x in all_matching_alleles if x[1] == highest_score]
        most_common_alleles = [x[0] for x in most_common]
        most_common_alleles.sort()
        return most_common_alleles, highest_score

    else:
        return '', ''


def flatten(list_of_lists):
    """
    :param list_of_lists: pretty well described in the name here
    :return: flattened list of items within parent lists
    """
    return [x for xs in list_of_lists for x in xs]


def infer(seq, digits, seqtype, hla_loci, data_dir, lower_len, higher_len, window_len):
    """
    :param seq: str, nucletide or amino acid to try to identify an HLA allele in
    :param digits: int, resolution of HLA allele number (2/4/6/8)
    :param seqtype: str, sequence type as per IMGTHLA (gen, nuc, prot)
    :param hla_loci: list of str, detailing IMGTLA pre-asterisk gene identifiers
    :param data_dir: str, path to hladl data directory
    :param lower_len: int, lower length limit of sequence allowed (in amino acids)
    :param higher_len: int, upper length limit of sequence allowed (in amino acids)
    :param window_len: int, length of sliding window to use for overlapping tags
    :return: 2 items: list of str of inferred alleles, and int of # of tags they matched
    """

    # Get the relevant data, and search the sequence
    hla_loci = hla_loci.split(',')
    trie, tags = get_trie_data(digits, seqtype, data_dir, hla_loci, lower_len, higher_len, window_len)

    # TODO add automatic sequence type checks?
    inferred_alleles, tag_count = infer_hla_allele(seq, trie, tags)

    return inferred_alleles, tag_count



# -*- coding: utf-8 -*-

from . import hladlfunctions as fxn


def seq(allele, digits, seqtype, mode, output_mode, data_dir):
    """
    :param allele: str, IMGTHLA format of HLA gene to produce
    :param digits: int, resolution of HLA allele number (2/4/6/8)
    :param seqtype: str, sequence type as per IMGTHLA (gen, nuc, prot)
    :param mode: str, mode describing sequences sought
    :param output_mode: str, describing how to output the results
    :param data_dir: str, path to hladl data directory
    :return: depends on mode: may return str of pulled allele, or save to file as requested
    """

    # Check input parameters
    fxn.check_digits(digits)
    if '*' not in allele:
        raise IOError("HLA allele ID should be the gene name minus 'HLA-': it requires the asterisk.")
    elif len(allele[allele.index('*'):]) != int(digits * 1.5):
        raise IOError("Requested allele length is not compatible with the requested digit resolution.")

    mode = mode.lower()
    if mode not in fxn.modes:
        raise IOError(f"Requested mode ({mode}) is not a recognised mode ({fxn.modes}).")

    if mode == 'ecd' and (seqtype == 'nuc' or allele[0] not in fxn.featured_mhci):
        raise IOError("The extracellular domain mode is currently only configured"
                      " to work with classical MHC-I protein sequences.")

    # Get the relevant data
    hla, data_file = fxn.get_data(digits, seqtype, data_dir)

    if allele not in hla:
        raise IOError("Allele not present in HLA dictionary.")
    else:
        hla_seq = hla[allele]

    # Trim off leader and trans-membrane/intracellular portions if requested
    if mode == 'ecd':
        hla_len = len(hla_seq)
        if hla_len < 350 or hla_len > 380:
            raise IOError(f"Unable to extract extracellular domain, sequence an unexpected length ({hla_len} AA).")
        hla_seq = hla_seq[24:308]
        # TODO make smarter? PWMs?

    if output_mode == 'stdout':
        return hla_seq
    elif output_mode == 'fasta':
        out_str = fxn.fastafy(allele, hla_seq)
        out_name = allele.replace('*', '').replace(':', '-') + '.fasta'
        with open(out_name, 'w') as out_file:
            out_file.write(out_str)



# -*- coding: utf-8 -*-

import textwrap
import os
import gzip
import json
from . import download as dl


def readfq(fastx_file):
    """
    readfq(file):Heng Li's Python implementation of his readfq function
    https://github.com/lh3/readfq/blob/master/readfq.py
    :param fastx_file: opened file containing fastq or fasta reads
    :yield: read id, read sequence, and (where available) read quality scores
    """

    last = None  # this is a buffer keeping the last unprocessed line
    while True:
        if not last:  # the first record or a record following a fastq
            for l in fastx_file:  # search for the start of the next record
                if l[0] in '>@':  # fasta/q header line
                    last = l[:-1]  # save this line
                    break

        if not last:
            break

        name, seqs, last = last[1:], [], None  # This version takes the whole line (post '>')
        for l in fastx_file:  # read the sequence
            if l[0] in '@+>':
                last = l[:-1]
                break
            seqs.append(l[:-1])

        if not last or last[0] != '+':  # this is a fasta record
            yield name, ''.join(seqs), None  # yield a fasta record
            if not last:
                break

        else:  # this is a fastq record
            sequence, leng, seqs = ''.join(seqs), 0, []
            for l in fastx_file:  # read the quality
                seqs.append(l[:-1])
                leng += len(l) - 1
                if leng >= len(sequence):  # have read enough quality
                    last = None
                    yield name, sequence, ''.join(seqs)  # yield a fastq record
                    break

            if last:  # reach EOF before reading enough quality
                yield name, sequence, None  # yield a fasta record instead
                break


def fastafy(gene, seq_line):
    """
    :param gene: Gene symbol, extracted from the read id
    :param seq_line: Total protein primary sequence, extracted from input FASTA/generated by in silico splicing
    :return: An output-compatible FASTA entry ready for writing to file
    """
    return ">" + gene + "\n" + textwrap.fill(seq_line, 60) + "\n"


def trim_gene(digits, full_gene):
    """
    :param digits: int, number of digits to give HLA resolution (2/4/6/8)
    :param full_gene: identifier of full gene/allele, as given in the second space-limited field of the IMGTHLA headers
    :return:
    """
    asterisk_x = full_gene.find('*')
    if asterisk_x == -1:
        raise IOError("Full allele identifier (minus 'HLA-') required, inclusive of asterisk.")
    else:
        return full_gene[:asterisk_x + int(1.5 * digits)]


def check_digits(requested_digits):
    if requested_digits not in [2, 4, 6, 8]:
        raise IOError("The only acceptable value for 'digits' are 2, 4, 6, and 8. ")


def get_data(digits, seqtype, data_dir):
    # TODO docstr

    # Check to see if the relevant data is present
    type_match = '_' + str(digits) + '_' + seqtype
    data_files = [x for x in os.listdir(data_dir) if x.endswith('.json.gz')
                  and type_match in x and 'tags' not in x]

    if not data_files:
        print("Necessary data not detected: downloading. ")
        dl.get_data(seqtype, digits, data_dir)
        data_files = [x for x in os.listdir(data_dir) if x.endswith('.json.gz') and type_match in x]

    # Use the most recent entry
    data_files.sort()
    recent = data_files[-1]
    hla = read_json(os.path.join(data_dir, recent))
    return hla, recent


def save_json(out_path, to_save, print2stdout=True):
    """
    :param out_path: str, path to save JSON to
    :param to_save: dict to save as JSON
    :param print2stdout: bool, detailing whether to print a confirmation after saving
    :return: nothing
    """
    with gzip.open(out_path, 'wt') as out_file:
        json.dump(to_save, out_file)

    if print2stdout:
        print('\tSaved to', out_path)


def read_json(in_path):
    """
    :param in_path: str, path to JSON file to read in
    :return: parsed JSON document
    """
    try:

        with gzip.open(in_path, 'rt') as in_file:
            return json.load(in_file)

    except Exception:
        raise IOError(f"Unable to read in JSON file '{in_path}'.")


modes = ['full', 'ecd']

# Details of the sequences and locations of the relevant domains, based on Uniprot examples
domains = {'signal': {'A': 'MAVMAPRTLLLLLSGALALTQTWA',  # https://www.uniprot.org/uniprotkb/P04439/
                      'B': 'MLVMAPRTVLLLLSAALALTETWA',  # https://www.uniprot.org/uniprotkb/P01889/
                      'C': 'MRVMAPRALLLLLSGGLALTETWA'},  # https://www.uniprot.org/uniprotkb/P10321/
           'transmemb': {'A': 'VGIIAGLVLLGAVITGAVVAAVMW',
                         'B': 'GIVAGLAVLAVVVIGAVVAAVMCR',
                         'C': 'MGIVAGLAVLVVLAVLGAVVTAMMC'}}

domain_locs = {'signal': {'A': '1-24',
                          'B': '1-24',
                          'C': '1-24'},
               'transmemb': {'A': '309-332',
                             'B': '310-333',
                             'C': '309-333'}}

featured_mhci = ['A', 'B', 'C', 'E']



# -*- coding: utf-8 -*-

import requests
import gzip
import os
import collections as coll
from . import hladlfunctions as fxn

data_url_base = 'https://raw.githubusercontent.com/ANHIG/IMGTHLA/refs/heads/Latest/fasta/hla_REPLACE.fasta'
release_url = 'https://raw.githubusercontent.com/ANHIG/IMGTHLA/refs/heads/Latest/release_version.txt'


def get_data(seqs, digits, data_dir):

    seqs = seqs.lower()
    if seqs not in ['nuc', 'prot', 'all']:
        raise IOError("Sequence type needs to be one of "
                      "'gen' (gDNA), 'nuc' (cDNA)', 'prot' (AA), or 'all' (get all 3). ")
    if seqs == 'all':
        seqs = ['nuc', 'prot']
    else:
        seqs = [seqs]

    fxn.check_digits(digits)

    # Grab release info first
    response = requests.get(release_url, headers={})
    if response.status_code == 200:
        release_dat = response.text.split('\n')
        release_line = [x for x in release_dat if 'version:' in x]
        if len(release_line) == 1:
            release = release_line[0].split(' ')[-1]
        else:
            raise IOError("Unable to identify release # from repo!")

    out_prefix = release + '_' + str(digits) + '_'

    print(f"Downloading reference data ('{seqs}'), at {digits} digit annotation, saving files to:")
    # Then grab the corresponding file(s)
    for seq in seqs:
        url = data_url_base.replace('REPLACE', seq)

        response = requests.get(url, headers={})
        if response.status_code == 200:
            data = response.content.decode()

            raw_path = os.path.join(data_dir, out_prefix + seq + '.fasta.gz')
            with gzip.open(raw_path, 'wt') as ref_file:
                ref_file.write(data)

            # Then read that file in
            hla_dict = coll.defaultdict(str)
            with gzip.open(raw_path, 'rt') as ref_file:
                for fa_header, fa_seq, fa_null in fxn.readfq(ref_file):
                    full_gene = fa_header.split(' ')[1]
                    gene = fxn.trim_gene(digits, full_gene)

                    # Only take one sequence per allele at the requested resolution
                    if gene not in hla_dict:
                        hla_dict[gene] = fa_seq

            # Then write that out that dict for later use
            processed_path = os.path.join(data_dir, out_prefix + seq + '.json.gz')
            fxn.save_json(processed_path, hla_dict)

        else:
            raise ConnectionError("Unable to reach the ANHIG/IMGTHLA Github repo!")

    print("Successfully completed hladl init!")
